import httpx
import jwt

from ..errors import (
    JWTAudienceError,
    JWTExpiredError,
    JWTPurposeError,
    JWTScopeError,
    JWTError,
)


def get_keys_url(issuer):
    # Prefer OIDC discovery doc, but fall back on Fence-specific /jwt/keys for backwards compatibility
    openid_cfg_path = "/".join(
        [issuer.strip("/"), ".well-known", "openid-configuration"]
    )
    try:
        jwks_uri = httpx.get(openid_cfg_path).json().get("jwks_uri", "")
        return jwks_uri
    except:
        return "/".join([issuer.strip("/"), "jwt", "keys"])


def get_kid(encoded_token):
    try:
        return jwt.get_unverified_header(encoded_token).get("kid")
    except jwt.InvalidTokenError as e:
        raise JWTError(e)


def get_iss(encoded_token):
    try:
        return jwt.decode(encoded_token, verify=False).get("iss")
    except jwt.InvalidTokenError as e:
        raise JWTError(e)


def validate_purpose(claims, pur):
    """
    Check that the claims from a JWT have the expected purpose claim ``pur``.

    Args:
        claims (dict): claims from token
        pur (str): expected purpose

    Return:
        None

    Raises:
        JWTPurposeError:
            if the claims do not contain a purpose claim or if it doesn't match
            the expected value
    """
    if "pur" not in claims:
        raise JWTPurposeError("claims missing `pur` claim")
    if claims["pur"] != pur:
        raise JWTPurposeError(
            "claims have incorrect purpose: expected {}, got {}".format(
                pur, claims["pur"]
            )
        )


def validate_jwt(encoded_token, public_key, aud, scope, issuers, options={}):
    """
    Validate the encoded JWT ``encoded_token``, which must satisfy the
    scopes ``scope``.

    This is just a slightly lower-level function to decode the token and
    perform the most basic checks on the token.

    - Decode JWT using public key; PyJWT will fail if iat or exp fields are
      invalid
    - PyJWT will also fail if the aud field is present in the JWT but no
      ``aud`` arg is passed, or if the ``aud`` arg does not match one of
      the items in the token aud field
    - Check issuers: token iss field must match one of the items in the
      ``issuers`` arg
    - Check scopes: token scopes must be a superset of required scopes
      (the ``scope`` argument); fail if not satisfied

    Args:
        encoded_token (str): encoded JWT
        public_key (str): public key to validate the JWT signature
        aud (Optional[str]):
          audience with which the app identifies, usually an OIDC
          client id, which the JWT will be expected to include in its ``aud``
          claim. Optional; if no ``aud`` argument given, then the JWT must
          not have an ``aud`` claim, or validation will fail.
        scope (Optional[Iterable[str]]):
          set of scopes, each of which the JWT must satisfy in its
          ``scope`` claim. Optional.
        issuers (list or set): allowed issuers whitelist
        options (Optional[dict]): options to pass through to pyjwt's decode

    Return:
        dict: the decoded and validated JWT

    Raises:
        ValueError: if receiving an incorrectly-typed argument
        JWTExpiredError: if token is expired
        JWTAudienceError: if aud validation fails
        JWTScopeError: if scope validation fails
        JWTError: if some other token validation step fails
    """

    # Typecheck arguments.
    if not isinstance(aud, str) and not aud is None:
        raise ValueError(
            "aud must be string or None. Instead received aud of type {}".format(
                type(aud)
            )
        )
    if not isinstance(scope, set) and not isinstance(scope, list) and not scope is None:
        raise ValueError(
            "scope must be set or list or None. Instead received scope of type {}".format(
                type(scope)
            )
        )
    if not isinstance(issuers, set) and not isinstance(issuers, list):
        raise ValueError(
            "issuers must be set or list. Instead received issuers of type {}".format(
                type(issuers)
            )
        )

    try:
        token = jwt.decode(
            encoded_token,
            key=public_key,
            algorithms=["RS256"],
            audience=aud,
            options=options,
        )
    except jwt.InvalidAudienceError as e:
        raise JWTAudienceError(e)
    except jwt.ExpiredSignatureError as e:
        raise JWTExpiredError(e)
    except jwt.InvalidTokenError as e:
        raise JWTError(e)

    # PyJWT validates iat, exp, and aud fields; everything else
    # must happen here.

    # iss
    # Check that the issuer of the token has the expected hostname.
    if token["iss"] not in issuers:
        msg = "invalid issuer {}; expected: {}".format(token["iss"], issuers)
        raise JWTError(msg)

    # scope
    # Check that if scope arg was non-empty then the token includes each given scope in its scope claim
    if scope:
        token_scopes = token.get("scope", [])
        if isinstance(token_scopes, str):
            token_scopes = token_scopes.split()
        if not isinstance(token_scopes, list):
            raise JWTError(
                "invalid format in scope claim: {}; expected string or list".format(
                    token["scopes"]
                )
            )
        missing_scopes = set(scope) - set(token_scopes)
        if missing_scopes:
            raise JWTScopeError(
                "token is missing required scopes: " + str(missing_scopes)
            )

    return token



# re-export for compatibility
try:
    from authutils.token.validate import (
        current_token,
        set_current_token,
        store_session_token,
        get_session_token,
        get_jwt_token,
    )
except ImportError:
    current_token = (
        set_current_token
    ) = store_session_token = get_session_token = get_jwt_token = None

__all__ = [
    "current_token",
    "get_session_token",
    "set_current_token",
    "store_session_token",
    "get_jwt_token",
]



from authutils.errors import AuthError

try:
    from authutils.user import CurrentUser, set_global_user
except ImportError:
    CurrentUser = set_global_user = None

__all__ = ["AuthError", "CurrentUser", "ROLES", "set_global_user"]

ROLES = dict(
    ADMIN="admin",
    CREATE="create",
    DELETE="delete",
    DOWNLOAD="download",
    GENERAL="_member_",
    READ="read",
    RELEASE="release",
    UPDATE="update",
)



from asyncio import Future, get_event_loop
from collections import OrderedDict

import httpx
from fastapi import Security, HTTPException
from fastapi.security import HTTPBearer, HTTPAuthorizationCredentials
from starlette.status import HTTP_403_FORBIDDEN

from . import core
from ..errors import JWTError, AuthError

bearer = HTTPBearer()

_jwt_public_keys = {}


def access_token(
    *scopes, audience=None, issuer=None, allowed_issuers=None, purpose=None
):
    """
    Validate and return the JWT bearer token in HTTP header::

        from fastapi import Depends

        @app.get("/whoami")
        def whoami(token=Depends(access_token("user", "openapi", purpose="access"))):
            return token["iss"]

    Args:
        *scopes: Required, all must occur in ``scope``.
        audience: Optional; if provided, JWT validation will require that the token's
          ``aud`` value contains the arg value; if not provided, validation will require
          that the token not have an aud field.
        issuer: Optional; force to use this issuer to validate the token if provided.
        allowed_issuers: Optional allowed issuers whitelist, default: allow all.
        purpose: Optional, must match ``pur`` if provided.

    Returns:
        Decoded JWT claims as a :class:`dict`.
    """

    if not scopes:
        raise ValueError("Missing parameter: scopes")
    scopes = set(scopes)
    if not allowed_issuers and issuer:
        allowed_issuers = [issuer]

    async def getter(token: HTTPAuthorizationCredentials = Security(bearer)):
        nonlocal issuer, allowed_issuers
        assert token.scheme.lower() == "bearer"
        token = token.credentials
        loop = get_event_loop()

        # get kid and issuer
        try:
            kid = await loop.run_in_executor(None, core.get_kid, token)
            if issuer is None:
                issuer = await loop.run_in_executor(None, core.get_iss, token)
        except JWTError as e:
            raise HTTPException(
                status_code=HTTP_403_FORBIDDEN, detail="Bad bearer token: " + str(e)
            )
        if not allowed_issuers:
            allowed_issuers = [issuer]
        if issuer not in allowed_issuers:
            raise HTTPException(
                status_code=HTTP_403_FORBIDDEN,
                detail="Bad bearer token: issuer is not allowed: " + issuer,
            )

        # get public key from cache, or fetch from issuer
        pub_keys = _jwt_public_keys.get(issuer)
        if not pub_keys:
            pub_keys = _jwt_public_keys[issuer] = Future()
            try:
                async with httpx.AsyncClient() as client:
                    resp = await client.get(core.get_keys_url(issuer))
                    resp.raise_for_status()
                    pub_keys.set_result(OrderedDict(resp.json()["keys"]))
            except Exception as e:
                _jwt_public_keys.pop(issuer)
                pub_keys.set_exception(
                    HTTPException(
                        status_code=HTTP_403_FORBIDDEN,
                        detail="Cannot fetch pubkey from issuer {}: {}".format(
                            issuer, str(e)
                        ),
                    )
                )
        pub_keys = await pub_keys
        pub_key = pub_keys.get(kid)
        if not pub_key:
            raise HTTPException(
                status_code=HTTP_403_FORBIDDEN,
                detail="Bad bearer token: kid not found in issuer: " + issuer,
            )

        # decode and validate the token
        try:
            claims = await loop.run_in_executor(
                None,
                core.validate_jwt,
                token,
                pub_key,
                audience,
                scopes,
                allowed_issuers,
            )

            if purpose:
                core.validate_purpose(claims, purpose)
        except AuthError as e:
            raise HTTPException(
                status_code=HTTP_403_FORBIDDEN, detail="Bad bearer token: " + str(e)
            )

        return claims

    return getter



# -*- coding: utf-8 -*-
"""
Defines functionality to check case existence in dbGaP.
"""

import httpx
import re

import xmltodict
from cdislogging import get_logger
from datamodelutils import models
from xml.parsers.expat import ExpatError
from cdiserrors import InternalError, UserError

PHSID_REGEX = re.compile("(phs\d+.v)(\d)(.*)")

COMPLETE_STATE = ["released", "completed_by_gpa"]


class dbGaPXReferencer(object):

    #: The url from which to pull telemetry reports for project with given
    #: accession number (to be formatted in)
    DEFAULT_URL = (
        "https://www.ncbi.nlm.nih.gov/projects/gap/cgi-bin/"
        "GetSampleStatus.cgi?study_id={phsid}&rettype=xml"
    )

    def __init__(self, db, logger=None, proxies={}):
        """Instantiate a class to crossvalidate entity existence in dbGaP.

        """
        self._cached_telemetry_xmls = {
            # "phsid": "telemetry xml"
        }

        self.db = db
        self.proxies = proxies
        self.logger = logger or get_logger("dbGapXReferencer", log_level="info")
        self.logger.info("Creating new dbGaP Cross Referencer")

    def request_telemetry_report(self, phsid):
        """Makes a web request to :param:`url` for the telemetry report for
        the given :param:`phsid`

        :param str phsid:
            The accession number of the project to get the telemetry XML
            report for.
        :param str url:
            *optional* Specify the dbGaP url
        :raises gdcapi.errors.InternalError:
            Will try and raise a descriptive InternalError if something
            goes wrong.
        :returns: A dict parsed from the XML

        """

        xml = self.get_xml(phsid)

        # use previous version of telemetry report
        # if current one is not released

        if xml["DbGap"]["Study"]["@registration_status"] not in COMPLETE_STATE:
            current_version = xml["DbGap"]["Study"]["@accession"]
            match = PHSID_REGEX.match(current_version)
            if not match:
                raise InternalError(
                    "Unable to cross reference cases with dbGaP. "
                    "Invalid accession number {} in telemery report from dbGap".format(
                        phsid
                    )
                )
            else:
                previous_version = (
                    match.group(1) + str(int(match.group(2)) - 1) + match.group(3)
                )
                xml = self.get_xml(previous_version)
                if xml["DbGap"]["Study"]["@registration_status"] != "released":
                    raise InternalError(
                        "Unable to cross reference cases with dbGaP. "
                        "Last two versions of telemetry reports from dbGap "
                        "are not released"
                    )
                return xml
        else:
            return xml

    def get_xml(self, phsid):
        if phsid not in self._cached_telemetry_xmls:

            url = self.DEFAULT_URL.format(phsid=phsid)
            self.logger.info("Pulling telemetry report from {0}".format(url))

            # Request the XML
            with httpx.Client(proxies=self.proxies) as client:
                r = client.get(url)
            if r.status_code != 200:
                msg = (
                    "Unable to cross reference cases with dbGaP. "
                    "Either this project is not registered in dbGaP or we "
                    "were temporarily unable to communicate with dbGaP. "
                    "Please try again later."
                )
                self.logger.error(msg)
                raise InternalError(msg)

            # Parse the XML
            try:
                xml = xmltodict.parse(r.text)
                self._cached_telemetry_xmls[phsid] = xml

            except ExpatError as e:
                msg = (
                    "Unable to parse dbGaP telemetry report. " "Please try again later."
                )
                self.logger.exception(e)
                raise InternalError(msg)

        return self._cached_telemetry_xmls[phsid]

    def get_project(self, program_name, project_code):
        """Lookup the project node

        :param str program_name: the `Program.name` e.g. TCGA
        :param str project_code: the `Project.code` e.g.e BRCA
        :returns: :class:`Project` node

        """

        self.logger.info(
            "Looking up project {0}-{1}".format(program_name, project_code)
        )

        with self.db.session_scope():
            project = (
                self.db.nodes(models.Project)
                .props(code=project_code)
                .path("programs")
                .props(name=program_name)
                .scalar()
            )

            if not project:
                msg = "Unable to find project {} in database".format(project_code)
                self.logger.error(msg)
                raise InternalError(msg)

        return project

    def get_project_accession(self, project):
        """Return the project's accession number (phsid)

        :param project: the `Project` node
        :returns: String accession number

        """

        # Return the projects phsid, (default to the project's
        # program's phsid)
        return (
            project.dbgap_accession_number or project.programs[0].dbgap_accession_number
        )

    def get_project_dbgap_bypassed_cases(self, project):
        """Check if there is a list of bypassed cases associated with the
        project.

        :param project: the `Project` node
        :returns: :class:`Project` a list of submitter_id strings

        """

        return project.sysan.get("dbgap_bypassed_cases") or []

    def get_registered_cases(self, project):
        """Gets the submitter_id of all the cases in latest telemetry report.
        """
        phsid = self.get_project_accession(project)

        # Pull the telemetry report for the project
        telemetry = self.request_telemetry_report(phsid)

        # Parse our the sample ids
        samples = telemetry["DbGap"]["Study"]["SampleList"]["Sample"]
        return {s["@submitted_subject_id"] for s in samples}

    def case_exists(self, program_name, project_code, case_submitter_id):
        """Checks to see if case exists in latest telemetry report.

        :param str program_name: the `Program.name` e.g. TCGA
        :param str project_code: the `Project.code` e.g.e BRCA
        :returns: :class:`bool`

        :raises:
            :class:`gdcapi.errors.InternalError` if the project isn't
            found

        """

        # Lookup the project node
        project = self.get_project(program_name, project_code)
        # Check against a local bypass list
        if case_submitter_id in self.get_project_dbgap_bypassed_cases(project):
            self.logger.warning(
                "Found case {} in local bypass list".format(case_submitter_id)
            )
            return True

        submitter_ids = self.get_registered_cases(project)
        return case_submitter_id in submitter_ids

    def assert_project_exists(self, project_code, phsid):
        url = self.DEFAULT_URL.format(phsid=phsid)
        self.logger.info("Pulling telemetry report from {0}".format(url))

        # Request the XML
        with httpx.Client(proxies=self.proxies) as client:
            r = client.get(url)
        if r.status_code == 400:
            msg = "Project appears not to exist in dbGaP."
            raise UserError(msg)

        # Parse the XML
        try:
            xml = xmltodict.parse(r.text)
        except ExpatError as e:
            msg = "Unable to parse dbGaP telemetry report. " "Please try again later."
            self.logger.exception(e)
            raise InternalError(msg)



from authlib.common.errors import AuthlibBaseError
import flask
from flask import current_app

from authutils.errors import AuthError
from authutils.token import store_session_token


def client_do_authorize():
    redirect_uri = current_app.oauth_client.session.redirect_uri
    mismatched_state = (
        "state" not in flask.request.args
        or "state" not in flask.session
        or flask.request.args["state"] != flask.session.pop("state")
    )
    if mismatched_state:
        raise AuthError("could not authorize; state did not match across auth requests")
    try:
        token = current_app.oauth_client.fetch_access_token(
            redirect_uri, **flask.request.args.to_dict()
        )
        store_session_token(token["access_token"])
        return token
    except KeyError as e:
        raise AuthError("error in token response: {}".format(token))
    except AuthlibBaseError as e:
        raise AuthError(str(e))



"""
NOTE: the key fixtures depend on ``_hazmat_rsa_private_key`` and
``_hazmat_rsa_private_key_2``, so these must be imported as well (even if not
used) in order for the fixtures to work.
"""

from authutils.testing.fixtures.keys import (
    _hazmat_rsa_private_key,
    _hazmat_rsa_private_key_2,
    rsa_private_key,
    rsa_private_key_2,
    rsa_public_key,
    rsa_public_key_2,
)



from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
import pytest


def _new_hazmat_rsa_private_key():
    """
    Return:
        cryptography.hazmat.backends.openssl.rsa._RSAPrivateKey
    """
    return rsa.generate_private_key(
        public_exponent=65537, key_size=2048, backend=default_backend()
    )


@pytest.fixture(scope="session")
def _hazmat_rsa_private_key():
    """
    Return:
        cryptography.hazmat.backends.openssl.rsa._RSAPrivateKey
    """
    return _new_hazmat_rsa_private_key()


@pytest.fixture(scope="session")
def _hazmat_rsa_private_key_2():
    """
    Make a different private key than the first one.

    Return:
        cryptography.hazmat.backends.openssl.rsa._RSAPrivateKey
    """
    return _new_hazmat_rsa_private_key()


@pytest.fixture(scope="session")
def rsa_private_key(_hazmat_rsa_private_key):
    """
    Return a string of an RSA private key in PEM format.

    Args:
        _hazmat_rsa_private_key: fixture

    Return:
        str: RSA private key

    Example:

        .. code-block:: python

            \"\"\"
            -----BEGIN RSA PRIVATE KEY-----
            VbX7OiPS...
            ...
            ...ZHxIKy2+
            -----END RSA PRIVATE KEY-----
            \"\"\"
    """
    return _hazmat_rsa_private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption(),
    ).decode("utf-8")


@pytest.fixture(scope="session")
def rsa_private_key_2(_hazmat_rsa_private_key_2):
    """
    Return a (second, different) string of an RSA private key in PEM format.

    Args:
        _hazmat_rsa_private_key: fixture

    Return:
        str: RSA private key

    Example:

        .. code-block:: python

            \"\"\"
            -----BEGIN RSA PRIVATE KEY-----
            VbX7OiPS...
            ...
            ...ZHxIKy2+
            -----END RSA PRIVATE KEY-----
            \"\"\"
    """
    return _hazmat_rsa_private_key_2.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption(),
    ).decode("utf-8")


@pytest.fixture(scope="session")
def rsa_public_key(_hazmat_rsa_private_key):
    """
    Return a string of an RSA public key in PEM format. The key returned from
    this function matches the private key returned by ``rsa_private_key``.

    Args:
        _hazmat_rsa_private_key: fixture

    Return:
        str: RSA public key (including line breaks)

    Example:

        .. code-block:: python

            \"\"\"
            -----BEGIN PUBLIC KEY-----
            MIIBIjAN...
            ...
            ...BCgKCAQE
            -----END PUBLIC KEY-----
            \"\"\"
    """
    return (
        _hazmat_rsa_private_key.public_key()
        .public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo,
        )
        .decode("utf-8")
    )


@pytest.fixture(scope="session")
def rsa_public_key_2(_hazmat_rsa_private_key_2):
    """
    Return a string of a (second, different) RSA public key in PEM format. The
    key returned from this function matches the private key returned by
    ``rsa_private_key_2``.

    Args:
        _hazmat_rsa_private_key: fixture

    Return:
        str: RSA public key (including line breaks)

    Example:

        .. code-block:: python

            \"\"\"
            -----BEGIN PUBLIC KEY-----
            MIIBIjAN...
            ...
            ...BCgKCAQE
            -----END PUBLIC KEY-----
            \"\"\"
    """
    return (
        _hazmat_rsa_private_key_2.public_key()
        .public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo,
        )
        .decode("utf-8")
    )



from cdiserrors import AuthNError as CDISAuthNError


class AuthError(CDISAuthNError):

    pass


class JWTError(AuthError):

    pass


class JWTExpiredError(AuthError):

    pass


class JWTPurposeError(JWTError):

    pass


class JWTAudienceError(JWTError):

    pass


class JWTScopeError(JWTError):

    pass



# pylint: disable=protected-access
"""
Define functions for validating a JWT and tracking the current token and claims
from a request.
"""

import flask
import functools
from cdislogging import get_logger
from werkzeug.local import LocalProxy

from . import core
from .keys import get_public_key_for_token
from ..errors import JWTError

#: Proxy for the current token, which gets assigned to in
#: ``require_auth_header``, so any function using that as a decorator can then
#: use ``current_token`` to get the claims for the token.
#:
#: (Note that this should really be called `set_current_claims`, since we're
#: using the claims here and not the raw token, which *is* what gets used for
#: `store_session_token`/`get_session_token`.)
#:
#: Anything that needs to check a JWT should try to use this, and set it if it
#: wasn't already, to avoid redundant validation.
current_token = LocalProxy(lambda: getattr(flask.g, "_current_token", None))


def set_current_token(token):
    flask.g._current_token = token


def store_session_token(token):
    flask.session["_authutils_access_token"] = token


def get_jwt_token():
    """
    Get the current JWT (in encoded form), from a bearer auth header.

    Requires flask request context.

    Return:
        str: encoded token
    """
    auth_header = flask.request.headers.get("Authorization")
    token = None
    if auth_header:
        items = auth_header.split(" ")
        if len(items) == 2 and items[0].lower() == "bearer":
            token = items[1]
    return token


def get_session_token():
    """
    Get the current JWT (in encoded form), from either a bearer auth header or
    the `_authutils_access_token` in the flask session.

    Requires flask request context and a matching Session interface.

    Return:
        str: encoded token
    """
    return get_jwt_token() or flask.session.get("_authutils_access_token")


def validate_jwt(
    encoded_token,
    aud=None,
    scope=None,
    purpose="access",
    issuers=None,
    public_key=None,
    attempt_refresh=True,
    logger=None,
    options={},
):
    """
    Validate a JWT and return the claims.

    Args:
        encoded_token (str): the base64 encoding of the token
        aud (Optional[str]):
            audience as which the app identifies, which the JWT will be
            expected to include in its ``aud`` claim.
            Optional; will default to issuer from flask.current_app.config
            if available (either BASE_URL or USER_API).
            To skip aud validation, pass the following in the options arg:
              options={"verify_aud": False}
        scope (Optional[Iterable[str]]):
            scopes that the token must satisfy
        purpose (Optional[str]):
            which purpose the token is supposed to be used for (access,
            refresh, or id)
        issuers (Iterable[str]): list of allowed token issuers
        public_key (Optional[str]): public key to vaidate JWT with
        attempt_refresh (Optional[bool]):
            whether to attempt refresh of public keys if not found in cache
        options (Optional[dict]): options to pass through to pyjwt's decode

    Return:
        dict: dictionary of claims from the validated JWT

    Raises:
        JWTError:
            if auth header is missing, decoding fails, or the JWT fails to
            satisfy any expectation
    """
    logger = logger or get_logger(__name__, log_level="info")

    if not issuers:
        issuers = []
        for config_var in ["OIDC_ISSUER", "USER_API", "BASE_URL"]:
            value = flask.current_app.config.get(config_var)
            if value:
                issuers.append(value)

    # Can't set arg default to config[x] in fn def, so doing it this way.
    if aud is None:
        aud = flask.current_app.config.get("BASE_URL")
    # Some Gen3 apps use BASE_URL and some use USER_API, so fall back on USER_API
    if aud is None:
        aud = flask.current_app.config.get("USER_API")

    if public_key is None:
        public_key = get_public_key_for_token(
            encoded_token, attempt_refresh=attempt_refresh, logger=logger
        )

    claims = core.validate_jwt(encoded_token, public_key, aud, scope, issuers, options)
    if purpose:
        core.validate_purpose(claims, purpose)
    return claims


def validate_request(scope={}, audience=None, purpose="access", logger=None):
    """
    Validate a ``flask.request`` by checking the JWT contained in the request
    headers.
    """
    logger = logger or get_logger(__name__, log_level="info")
    # Get token from the headers.
    try:
        encoded_token = flask.request.headers["Authorization"].split(" ")[1]
    except IndexError:
        raise JWTError("could not parse authorization header")
    except KeyError:
        raise JWTError("no authorization header provided")

    # Pass token to ``validate_jwt``.
    return validate_jwt(
        encoded_token,
        aud=audience,
        scope=scope,
        purpose=purpose,
        logger=logger,
    )


def require_auth_header(scope={}, audience=None, purpose=None, logger=None):
    """
    Return a decorator which adds request validation to check the given
    scopes, audience and purpose (all optional).
    """
    logger = logger or get_logger(__name__, log_level="info")

    def decorator(f):
        """
        Decorate the given function to check for a valid JWT header.
        """

        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            """
            Wrap a function to first validate the request header token.

            Assign the claims from the token to ``flask.g._current_token`` so
            the code inside the function can use the ``LocalProxy`` for the
            token (see top of this file).
            """
            set_current_token(
                validate_request(
                    scope=scope, audience=audience, purpose=purpose, logger=logger
                )
            )
            return f(*args, **kwargs)

        return wrapper

    return decorator



"""authutils.globals

global variables for auth
"""


ROLES = {
    "ADMIN": "admin",
    "CREATE": "create",
    "DELETE": "delete",
    "DOWNLOAD": "download",
    "GENERAL": "_member_",
    "READ": "read",
    "RELEASE": "release",
    "UPDATE": "update",
}

MEMBER_DOWNLOADABLE_STATES = ["submitted", "processing", "processed"]

SUBMITTER_DOWNLOADABLE_STATES = [
    "uploaded",
    "validating",
    "validated",
    "error",
    "submitted",
    "processing",
    "processed",
]



import functools
import json

from cached_property import cached_property
from cdiserrors import AuthZError
import flask
from werkzeug.local import LocalProxy

from authutils.errors import AuthError
from authutils.token.validate import set_current_token, validate_request


def set_current_user(**kwargs):
    default_expected_audience = flask.current_app.config.get("USER_API")
    # Gen3 services use both USER_API and BASE_URL
    if not default_expected_audience:
        default_expected_audience = flask.current_app.config.get("BASE_URL")

    # If not already passed an aud to expect, default to the application's url
    kwargs.setdefault("jwt_kwargs", {}).setdefault(
        "audience", default_expected_audience
    )

    flask.g.user = CurrentUser(**kwargs)
    set_current_token(flask.g.user._claims)
    return flask.g.user


# Proxy for the current user.
#
# Other modules importing authutils can import ``current_user`` from here,
# which will use ``_get_or_set_current_user`` to look up the user.
current_user = LocalProxy(set_current_user)


class CurrentUser(object):
    """
    Information about the user which issued a request.

    Args:
        _claims (dict): claims from the user's token (if validated already)
        jwt_kwargs (dict): keyword arguments to pass to ``validate_request``

    Attributes:
        _claims (dict): dictionary of claims from user token
        id (str): unique ID for the user
        username (str): user's username, according to token
        projects (Dict[str, List[str]): mapping of project IDs to roles
        is_admin (bool): whether the user has admin privileges
    """

    def __init__(self, claims=None, jwt_kwargs=None):
        jwt_kwargs = jwt_kwargs or {}
        if "scope" not in jwt_kwargs:
            jwt_kwargs["scope"] = {"openid"}
        self._claims = claims or validate_request(**jwt_kwargs)
        self.id = self._claims["sub"]
        self.username = self._get_user_info("name")
        self.projects = self._get_user_info("projects", default={})

    def __str__(self):
        str_out = {"id": self.id, "username": self.username, "is_admin": self.is_admin}
        return json.dumps(str_out)

    def _get_user_info(self, field, default=None):
        return self._claims.get("context", {}).get("user", {}).get(field, default)

    @cached_property
    def is_admin(self):
        """
        Indicate whether the current user has admin privileges.

        Return:
            bool: whether user is admin
        """
        # Try to just use the user context from the claims. If that doesn't
        # have the ``is_admin`` field then use the database lookup.
        return bool(self._get_user_info("is_admin"))

    def require_admin(self):
        """
        Raise an error if this user doesn't have admin privileges.
        """
        if not self.is_admin:
            raise AuthZError("user ({}) does not have admin privileges".format(self.id))

    def get_project_ids(self, role="_member_"):
        """
        Return a list of projects for which the user has this role.
        """
        return [project for project, roles in self.projects.items() if role in roles]


def set_global_user(**decorator_kwargs):
    """
    Wrap a Flask blueprint view function to set the global user
    ``flask.g.user`` to an instance of ``CurrentUser``, according to the
    information from the JWT in the request headers. The validation will also
    set the current token.

    This requires a flask application and request context.
    """

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            set_current_user(**decorator_kwargs)
            return func(*args, **kwargs)

        return wrapper

    return decorator



"""
Provide a basic set of endpoints for an application to implement OAuth client
functionality.

These endpoints assume that the ``current_app`` has already been configured
with an OAuth client instance from the ``authlib`` package as follows:

.. code-block:: python

    from authutils.oauth2.client import OAuthClient
    from service.api import app

    app.oauth_client = OAuthClient(
        'client-id',
        client_secret='...',
        api_base_url='https://api.auth.net/',
        access_token_url='https://auth.net/oauth/token',
        authorize_url='https://auth.net/oauth/authorize',
        client_kwargs={
            'scope': 'openid data user',
            'redirect_uri': 'https://service.net/authorize',
        },
    )

(NOTE the scopes are space-separated.)
"""

from urllib.parse import urljoin

from cdiserrors import APIError
import flask
from flask import current_app

import authutils.oauth2.client.authorize


blueprint = flask.Blueprint("oauth", __name__)


@blueprint.route("/authorization_url", methods=["GET"])
def get_authorization_url():
    """
    Provide a redirect to the authorization endpoint from the OP.
    """
    # This will be the value that was put in the ``client_kwargs`` in config.
    redirect_uri = current_app.oauth_client.session.redirect_uri
    # Get the authorization URL and the random state; save the state to check
    # later, and return the URL.
    authorization_url, state = current_app.oauth_client.generate_authorize_redirect(
        redirect_uri
    )
    flask.session["state"] = state
    return authorization_url


@blueprint.route("/authorize", methods=["GET"])
def do_authorize():
    """
    Send a token request to the OP.
    """
    authutils.oauth2.client.authorize.client_do_authorize()
    return "", 204


@blueprint.route("/logout", methods=["GET"])
def logout_oauth():
    """
    Log out the user.

    To accomplish this, just revoke the refresh token if provided.
    """
    url = urljoin(current_app.config.get("USER_API"), "/oauth2/revoke")
    token = flask.request.form.get("token")
    try:
        current_app.oauth_client.session.revoke_token(url, token)
    except APIError as e:
        msg = "could not log out, failed to revoke token: {}".format(e.message)
        return msg, 400
    return "", 204



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['authutils',
 'authutils.oauth2',
 'authutils.oauth2.client',
 'authutils.testing',
 'authutils.testing.fixtures',
 'authutils.token']

package_data = \
{'': ['*']}

install_requires = \
['authlib>=0.11,<1.0',
 'cached-property>=1.4,<2.0',
 'cdiserrors<2.0.0',
 'httpx>=0.12.1,<1.0.0',
 'pyjwt[crypto]>=1.5,<2.0',
 'xmltodict>=0.9,<1.0']

extras_require = \
{'fastapi': ['fastapi>=0.54.1,<0.55.0'], 'flask': ['Flask>=0.10.1']}

setup_kwargs = {
    'name': 'authutils',
    'version': '6.0.2',
    'description': 'Gen3 auth utility functions',
    'long_description': None,
    'author': 'CTDS UChicago',
    'author_email': 'cdis@uchicago.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



"""
Define functions for updating the public keys associated with certain token
issuers and retrieving the public key which can be used to verify a given JWT.

The public keys should be stored on the flask app in a `jwt_public_keys`
attribute, which will be a dictionary mapping issuer URLs (`iss` in a JWT) to
ordered dictionaries mapping key IDs to public key strings.

For example:

.. code-block:: python

    flask.current_app.jwt_public_keys == {
        'http://some-gen3-stack.net/user': OrderedDict([
            'key-01': '-----BEGIN PUBLIC KEY-----...',
            'key-02': '-----BEGIN PUBLIC KEY-----...',
        ]),
        'http://different-gen3-site.org/user': OrderedDict([
            'key-01': '-----BEGIN PUBLIC KEY-----...',
        ]),
    }
"""

import base64
import json
from collections import OrderedDict

from cdislogging import get_logger
import flask
import jwt
import httpx
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa


from authutils.errors import JWTError
from .core import get_keys_url, get_kid, get_iss


def refresh_jwt_public_keys(user_api=None, logger=None):
    """
    Update the public keys that the Flask app is currently using to validate
    JWTs.

    The get_keys_url helper function will prefer the user_api's
    .well-known/openid-configuration endpoint, but if no jwks_uri
    is found, will default to /jwt/keys.

    In the latter case, the response from ``/jwt/keys`` should look like this:

    .. code-block:: javascript

    {
        "keys": [
            [
                "key-id-01",
                "-----BEGIN PUBLIC KEY---- ... -----END PUBLIC KEY-----\n"
            ],
            [
                "key-id-02",
                "-----BEGIN PUBLIC KEY---- ... -----END PUBLIC KEY-----\n"
            ]
        ]
    }

    In either case, the keys are put into a dictionary and assigned to
    ``flask.current_app.jwt_public_keys`` with user_api as the key.
    Keys are serialized to PEM if not already.

    Args:
        user_api (Optional[str]):
            the URL of the user API to get the keys from; default to whatever
            the flask app is configured to use
        logger (Optional[Logger]):
            the logger; default to app's parent logger

    Return:
        None

    Side Effects:
        - Reassign ``flask.current_app.jwt_public_keys[user_api]`` to the keys obtained
          from ``get_jwt_public_keys``, as a dictionary.

    Raises:
        ValueError: if user_api is not provided or set in app config
    """
    logger = logger or get_logger(__name__, log_level="info")
    # First, make sure the app has a ``jwt_public_keys`` attribute set up.
    missing_public_keys = (
        not hasattr(flask.current_app, "jwt_public_keys")
        or not flask.current_app.jwt_public_keys
    )
    if missing_public_keys:
        flask.current_app.jwt_public_keys = {}
    user_api = user_api or flask.current_app.config.get("USER_API")
    if not user_api:
        raise ValueError("no URL(s) provided for user API")

    path = get_keys_url(user_api)
    try:
        jwt_public_keys = httpx.get(path).json()["keys"]
    except:
        raise JWTError(
            "Attempted to refresh public keys for {},"
            "but could not get keys from path {}.".format(user_api, path)
        )

    logger.info("Refreshing public key cache for issuer {}...".format(user_api))
    logger.debug(
        "Received public keys:\n{}".format(json.dumps(str(jwt_public_keys), indent=4))
    )

    issuer_public_keys = {}
    for key in jwt_public_keys:
        if "kty" in key and key["kty"] == "RSA":
            logger.debug(
                "Serializing RSA public key (kid: {}) to PEM format.".format(key["kid"])
            )
            # Decode public numbers https://tools.ietf.org/html/rfc7518#section-6.3.1
            n_padded_bytes = base64.urlsafe_b64decode(
                key["n"] + "=" * (4 - len(key["n"]) % 4)
            )
            e_padded_bytes = base64.urlsafe_b64decode(
                key["e"] + "=" * (4 - len(key["e"]) % 4)
            )
            n = int.from_bytes(n_padded_bytes, "big", signed=False)
            e = int.from_bytes(e_padded_bytes, "big", signed=False)
            # Serialize and encode public key--PyJWT decode/validation requires PEM
            rsa_public_key = rsa.RSAPublicNumbers(e, n).public_key(default_backend())
            public_bytes = rsa_public_key.public_bytes(
                serialization.Encoding.PEM,
                serialization.PublicFormat.SubjectPublicKeyInfo,
            )
            # Cache the encoded key by issuer
            issuer_public_keys[key["kid"]] = public_bytes
        else:
            logger.debug(
                "Key type (kty) is not 'RSA'; assuming PEM format. Skipping key serialization. (kid: {})".format(
                    key[0]
                )
            )
            issuer_public_keys[key[0]] = key[1]

    flask.current_app.jwt_public_keys.update({user_api: issuer_public_keys})
    logger.info("Done refreshing public key cache for issuer {}.".format(user_api))


def get_public_key(kid, iss=None, attempt_refresh=True, logger=None):
    """
    Given a key id ``kid``, get the public key from the flask app belonging to
    this key id. The key id is allowed to be None, in which case, use the the
    first key in the OrderedDict.

    - If current flask app is not holding public keys (ordered dictionary) or
      key id is in token headers and the key id does not appear in those public
      keys, refresh the public keys by calling ``refresh_jwt_public_keys()``
    - If key id is provided in the token headers:
      - If key id does not appear in public keys, fail
      - Use public key with this key id
    - If key id is not provided:
      - Use first public key in the ordered dictionary

    Args:
        kid (str): the key id
        attempt_refresh (bool):
            whether to try to refresh the public keys of the flask app if
            encountering a key id that does not exist in those keys; for fence
            itself this should be ``False``, and for other services it should
            be ``True``

    Return:
        str: the public key

    Side Effects:
        - From ``refresh_jwt_public_keys``: reassign
          ``flask.current_app.jwt_public_keys`` to the keys obtained from
          ``get_jwt_public_keys``.

    Raises:
        JWTValidationError:
            if the key id is provided and public key with that key id is found
    """

    iss = (
        iss
        or flask.current_app.config.get("OIDC_ISSUER")
        or flask.current_app.config["USER_API"]
    )
    logger = logger or get_logger(__name__, log_level="info")
    need_refresh = not hasattr(flask.current_app, "jwt_public_keys") or (
        kid and kid not in flask.current_app.jwt_public_keys.get(iss, {})
    )
    if need_refresh and attempt_refresh:
        refresh_jwt_public_keys(iss, logger=logger)
    elif need_refresh and not attempt_refresh:
        logger.warn(
            "Public key {} not cached, but application is not attempting refresh.".format(
                kid
            )
        )

    if iss not in flask.current_app.jwt_public_keys:
        raise JWTError("Public key for issuer {} not found.".format(iss))

    iss_public_keys = flask.current_app.jwt_public_keys[iss]
    try:
        return iss_public_keys[kid]
    except KeyError:
        raise JWTError("no key exists with given key id: {}".format(kid))


def get_public_key_for_token(encoded_token, attempt_refresh=True, logger=None):
    """
    Attempt to look up the public key which should be used to verify the token.

    Really just a thin wrapper around ``get_public_key`` which grabs the
    ``kid`` from the token headers and the ``iss`` from the token claims.

    Args:
        encoded_token (str): encoded JWT
        attempt_refresh (bool): whether to refresh public keys

    Return:
        str: public RSA key for token verification
    """
    logger = logger or get_logger(__name__, log_level="info")
    kid = get_kid(encoded_token)

    force_issuer = flask.current_app.config.get("FORCE_ISSUER")
    if force_issuer:
        iss = flask.current_app.config["USER_API"]
    else:
        iss = get_iss(encoded_token)
    return get_public_key(kid, iss=iss, attempt_refresh=attempt_refresh, logger=logger)
